// Parallel Mergesort
#include "stdio.h"

#define N 16	// 4*4
#define THREADS 4
#define BLOCKS 4

// This version only works for N = THREADS*BLOCKS
__global__ void sortBlocks(int *a)
{
	int i=2;
	__shared__ int temp[THREADS];
	while (i <= THREADS)
	{
	  if ((threadIdx.x % i)==0)
	  {
		int index1 = threadIdx.x + (blockIdx.x * blockDim.x);
		int endIndex1 = index1 + i/2;
		int index2 = endIndex1; 
		int endIndex2 = index2 + i/2;
		int targetIndex = threadIdx.x;
		int done = 0;
		while (!done)
		{
			if ((index1 == endIndex1) && (index2 < endIndex2))
				temp[targetIndex++] = a[index2++];
			else if ((index2 == endIndex2) && (index1 < endIndex1))
				temp[targetIndex++] = a[index1++];
			else if (a[index1] < a[index2])
				temp[targetIndex++] = a[index1++];
			else
				temp[targetIndex++] = a[index2++];
			if ((index1==endIndex1) && (index2==endIndex2))
				done = 1;
		}
	  }
	  __syncthreads();
	  a[threadIdx.x + (blockIdx.x*blockDim.x)] = temp[threadIdx.x];
	  __syncthreads();
	  i *= 2;
	}
}

__global__ void mergeBlocks(int *a, int *temp, int sortedsize)
{
	int id = blockIdx.x;
		
	int index1 = id * 2 * sortedsize;
	int endIndex1 = index1 + sortedsize;
	int index2 = endIndex1;
	int endIndex2 = index2 + sortedsize;
	int targetIndex = id * 2 * sortedsize;
	int done = 0;
	while (!done)
	{
		if ((index1 == endIndex1) && (index2 < endIndex2))
			temp[targetIndex++] = a[index2++];
		else if ((index2 == endIndex2) && (index1 < endIndex1))
			temp[targetIndex++] = a[index1++];
		else if (a[index1] < a[index2])
			temp[targetIndex++] = a[index1++];
		else
			temp[targetIndex++] = a[index2++];
		if ((index1==endIndex1) && (index2==endIndex2))
			done = 1;
	}
}

int main()
{
	int a[N];
	int *dev_a, *dev_temp;

	cudaMalloc((void **) &dev_a, N*sizeof(int));
	cudaMalloc((void **) &dev_temp, N*sizeof(int));

	// Fill array
	srand(time(NULL));
	for (int i = 0; i < N; i++)
	{
		int num = rand() % 100;
		a[i] = num;
		printf("%d ",a[i]);
	}
	printf("\n");

	// Copy data from host to device
	cudaMemcpy(dev_a, a, N*sizeof(int), cudaMemcpyHostToDevice);

	sortBlocks<<<BLOCKS,THREADS>>>(dev_a);
	cudaMemcpy(a, dev_a, N*sizeof(int), cudaMemcpyDeviceToHost);

	// Output Results
	for (int i = 0; i < N; i++)
	{
	  printf("%d ", a[i]);
	}
	printf("\n");

	int blocks = BLOCKS/2;
	int sortedsize = THREADS;
	while (blocks > 0)
	{
	  mergeBlocks<<<blocks,1>>>(dev_a, dev_temp, sortedsize);
	  cudaMemcpy(dev_a, dev_temp, N*sizeof(int), cudaMemcpyDeviceToDevice);
	  blocks /= 2;
	  sortedsize *= 2;
	}
	cudaMemcpy(a, dev_a, N*sizeof(int), cudaMemcpyDeviceToHost);
	
	// Output Results
	for (int i = 0; i < N; i++)
	{
	  printf("%d ", a[i]);
	}
	printf("\n");

	cudaFree(dev_a);
	cudaFree(dev_temp);
	return 0;
}
