// Larger array for mergesort
#include "stdio.h"

#define N 1048576
#define THREADS 512
#define BLOCKS 2048

__global__ void sortBlocks(int *a)
{
	int i=2;
	__shared__ int temp[THREADS];
	while (i <= THREADS)
	{
	  if ((threadIdx.x % i)==0)
	  {
		int index1 = threadIdx.x + (blockIdx.x * blockDim.x);
		int endIndex1 = index1 + i/2;
		int index2 = endIndex1; 
		int endIndex2 = index2 + i/2;
		int targetIndex = threadIdx.x;
		int done = 0;
		while (!done)
		{
			if ((index1 == endIndex1) && (index2 < endIndex2))
				temp[targetIndex++] = a[index2++];
			else if ((index2 == endIndex2) && (index1 < endIndex1))
				temp[targetIndex++] = a[index1++];
			else if (a[index1] < a[index2])
				temp[targetIndex++] = a[index1++];
			else
				temp[targetIndex++] = a[index2++];
			if ((index1==endIndex1) && (index2==endIndex2))
				done = 1;
		}
	  }
	  __syncthreads();
	  a[threadIdx.x + (blockIdx.x*blockDim.x)] = temp[threadIdx.x];
	  __syncthreads();
	  i *= 2;
	}
}

__global__ void mergeBlocks(int *a, int *temp, int sortedsize)
{
	int id = threadIdx.x + (blockIdx.x * blockDim.x);
		
	int index1 = id * 2 * sortedsize;
	int endIndex1 = index1 + sortedsize;
	int index2 = endIndex1;
	int endIndex2 = index2 + sortedsize;
	int targetIndex = id * 2 * sortedsize;
	int done = 0;
	while (!done)
	{
		if ((index1 == endIndex1) && (index2 < endIndex2))
			temp[targetIndex++] = a[index2++];
		else if ((index2 == endIndex2) && (index1 < endIndex1))
			temp[targetIndex++] = a[index1++];
		else if (a[index1] < a[index2])
			temp[targetIndex++] = a[index1++];
		else
			temp[targetIndex++] = a[index2++];
		if ((index1==endIndex1) && (index2==endIndex2))
			done = 1;
	}
}

void merge(int numbers[], int temp[], int left, int mid, int right)
{
  int i, left_end, num_elements, tmp_pos;
 
  left_end = mid - 1;
  tmp_pos = left;
  num_elements = right - left + 1;
 
  while ((left <= left_end) && (mid <= right))
  {
    if (numbers[left] <= numbers[mid])
    {
      temp[tmp_pos] = numbers[left];
      tmp_pos = tmp_pos + 1;
      left = left +1;
    }
    else
    {
      temp[tmp_pos] = numbers[mid];
      tmp_pos = tmp_pos + 1;
      mid = mid + 1;
    }
  }
 
  while (left <= left_end)
  {
    temp[tmp_pos] = numbers[left];
    left = left + 1;
    tmp_pos = tmp_pos + 1;
  }
  while (mid <= right)
  {
    temp[tmp_pos] = numbers[mid];
    mid = mid + 1;
    tmp_pos = tmp_pos + 1;
  }
 
  for (i=0; i <= num_elements; i++)
  {
    numbers[right] = temp[right];
    right = right - 1;
  }
}

void m_sort(int numbers[], int temp[], int left, int right)
{
  int mid;
 
  if (right > left)
  {
    mid = (right + left) / 2;
    m_sort(numbers, temp, left, mid);
    m_sort(numbers, temp, mid+1, right);
 
    merge(numbers, temp, left, mid+1, right);
  }
}

void mergesort(int numbers[], int temp[], int array_size)
{
  m_sort(numbers, temp, 0, array_size - 1);
}

int main()
{
	int a[N],*b,*temp;
	int *dev_a, *dev_temp;

	b = (int *) malloc(sizeof(int)*N);
	temp=(int *) malloc(sizeof(int)*N);
	cudaMalloc((void **) &dev_a, N*sizeof(int));
	cudaMalloc((void **) &dev_temp, N*sizeof(int));

	// Fill array
	srand(time(NULL));
	for (int i = 0; i < N; i++)
	{
		int num = rand();
		a[i] = num;
		b[i] = num;
//		printf("%d=%d ",i,a[i]);
	}
//	printf("\n");

	printf("Sorting using CPU mergesort...\n");
	mergesort(b,temp,N);
	printf("mergesort done.\n");

	// Copy data from host to device
	cudaMemcpy(dev_a, a, N*sizeof(int), cudaMemcpyHostToDevice);

	sortBlocks<<<BLOCKS,THREADS>>>(dev_a);
	cudaMemcpy(a, dev_a, N*sizeof(int), cudaMemcpyDeviceToHost);
	printf("Sorted blocks in parallel.\n");

	int blocks = BLOCKS / 2;
	int sortedsize = THREADS;
	while (blocks > 0)
	{
	  mergeBlocks<<<blocks,1>>>(dev_a, dev_temp, sortedsize);
	  cudaMemcpy(dev_a, dev_temp, N*sizeof(int), cudaMemcpyDeviceToDevice);
	  blocks /= 2;
	  sortedsize *= 2;
	}
	printf("Copying data.\n");
	cudaMemcpy(a, dev_a, N*sizeof(int), cudaMemcpyDeviceToHost);
	printf("Parallel mergesort complete.  Verifying results.\n");
	
	// Verify results
	int mismatch = 0;
	for (int i = 0; i <N; i++)
	{
		if (a[i]!=b[i])
		{
			mismatch=1;
			break;
		}
	}
	if (!mismatch)
		printf("Results match CPU mergesort!\n");
	else
		printf("Results do not match CPU mergesort!\n");

	cudaFree(dev_a);
	cudaFree(dev_temp);
	free(b);
	free(temp);
	return 0;
}
